""" Utility functions used by PXTL implementations.
"""

__all__= [
  'attrList', 'removeReparent', 'setDoctype', 'getNamespaces',
  'DictAsInstance', 'encodeText', 'decodeName', 'parsePseudoPIs',
  'ConditionalContext', 'checkFeatures', 'haveNestedScopes',
  'AbstractPXTLObject'
]

import types, sys, string, re, pxtl
from pxtl.constants import *
from pxtl.exceptions import *


# DOM utility functions
#
def attrList(nodes):
  """ Turn a live DOM NamedNodeMap into a static Python list, which can be
      used for potentially destructive iteration.
  """
  l= []
  for i in range(nodes.length):
    l.append(nodes.item(i))
  return l


def removeReparent(node):
  """ Replace a DOM node with its children, making sure all namespaces are
      the same on the child by copying declarations where necessary.
  """
  parentnss= None
  for i in range(node.attributes.length):
    attribute= node.attributes.item(i)
    if attribute.namespaceURI==NSNS:

      # Get a list of namespaces declared on ancestors (using one previously
      # worked out if this is not the first declaration). 
      #
      if parentnss is None:
        parentnss= []
        if node.parentNode is not None:
          parentnss= getNamespaces(node.parentNode)
      prefix= attribute.localName
      if attribute.localName=='xmlns':
        prefix= None

      # Ignore declarations that are redundant (declared on ancestor) or
      # overridden (declared on child). Copy the rest to children.
      #
      if parentnss.get(prefix)!=attribute.value:
        for child in list(node.childNodes):
          if child.nodeType==child.ELEMENT_NODE:
            if not child.hasAttributeNS(NSNS, attribute.localName):
              child.setAttributeNS(NSNS, attribute.name, attribute.value)

  # Replace node with a DocumentFragment of all its children.
  #
  frag= node.ownerDocument.createDocumentFragment()
  for child in list(node.childNodes):
    frag.appendChild(child)
  node.parentNode.replaceChild(frag, node)


def setDoctype(document, doctype):
  """ Try to set the doctype of a document to the specified object. This can't
      be done directly in DOM Level 2, and in Level 3 it's still not required,
      so direct replacement may not be possible with implementations other
      than pxdom. In this case, clone the document's content into a new
      document with the correct doctype.
  """
  current= document.doctype
  try:
    if current is None:
      if doctype is None:
        return document
      else:
        document.insertBefore(doctype, document.firstChild)
    else:
      if doctype is None:
        document.removeChild(current)
      else:
        document.replaceChild(doctype, current)
    return document
  except NonErrors:
    raise
  except Exception:

    # Failed in some way (why can't tell what, since DOMExceptions are not
    # bound to anything in the Python bindings - the exception could be
    # anything). Try the copying method instead.
    #
    imp= document.implementation
    newDocument= imp.createDocument(PXNS, 'px:none', doctype)
    frag= newDocument.createDocumentFragment()
    for child in list(document.childNodes):
      if child.nodeType!=child.DOCUMENT_TYPE_NODE:
        frag.appendChild(newDocument.importNode(child, True))
    newDocument.replaceChild(frag, newDocument.documentElement)

    # Try to copy any of the new properties from DOM Level 3 to the new
    # document; any of this could be unsupported and fail
    #
    try:
      newDocument.xmlVersion= document.xmlVersion
    except Exception:
      pass
    try:
      newDocument.xmlStandalone= document.xmlStandalone
    except Exception:
      pass
    try:
      newDocument.documentURI= document.documentURI
    except Exception:
      pass

    return newDocument


def getNamespaces(node):
  """ Get a mapping {prefix: uri} of all non-built-in namespaces active on an
      element.
  """
  nss= {}
  while node.nodeType==node.ELEMENT_NODE:
    for attr in attrList(node.attributes):
      if attr.namespaceURI==NSNS:
        prefix= attr.localName
        if attr.localName=='xmlns':
          prefix= None
        if not nss.has_key(prefix):
          nss[prefix]= attr.value
    node= node.parentNode
  return nss


# General-purpose utility functions.
#
class DictAsInstance:
  def __init__(self, d):
    self.__dict__= d

_encode_upar= re.compile('[^a-zA-Z0-9\\-\\._]')
_encode_jstr= re.compile('[\\\'"&<>-]')
_encode_cstr= _encode_jstr
_encode_name= re.compile('[^a-zA-Z0-9\\-\\.]')
_decode_name= re.compile('_..')

_hexdigits= '0123456789ABCDEF0123456789abcdef'
def _hexbyte(x):
  c= ord(x)
  return _hexdigits[c>>4]+_hexdigits[c&15]


def encodeText(text, coding):
  """ Return text in an escaped form (for URL parameters, JS literals etc.).
  """
  if coding in ('upar', 'name'):
    if type(text)==UnicodeType:
      text= text.encode('utf-8')
  if coding=='text':
    return text
  elif coding=='upar':
    return _encode_upar.sub(lambda m: '%'+_hexbyte(m.group(0)), text)
  elif coding=='jstr':
    return _encode_jstr.sub(lambda m: '\\x'+_hexbyte(m.group(0)), text)
  elif coding=='cstr':
    return _encode_cstr.sub(lambda m: '\\x'+_hexbyte(m.group(0))+' ', text)
  elif coding=='name':
    if text=='':
      return '___'
    text= _encode_name.sub(lambda m: '_'+_hexbyte(m.group(0)), text)
    if text[0] in '0123456789-.':
      text= '_'+_hexbyte(text[0])+text[1:]
    return text
  else:
    raise ValueError ("unknown coding '%s'" % coding)

def _decodeName_sub(m):
  c= m.group(0)[1:]
  if c[0] in _hexdigits and c[1] in _hexdigits:
    return chr(eval('0x'+c))
  return ''
def decodeName(text):
  text= _decode_name.sub(_decodeName_sub, text)
  if UnicodeType is None:
    return text
  return unicode(text, 'utf-8')


_ws= [' ', '\n', '\t']
def parsePseudoPIs(attr):
  """ Break an attribute value up into text and pseudo-PI sections
      (target, data).
  """
  v= attr.value
  parts= []
  ix= 0
  while True:
    ox= string.find(v, '{?', ix)
    if ox==-1:
      if ix<len(v):
        parts.append(v[ix:])
      break
    if ox>ix:
      parts.append(v[ix:ox])
    cx= string.find(v, '?}', ox+1)
    if cx==-1:
      raise PXTLPseudoPIOpenError(attr, ox)
    if cx==ox+1:
      parts.append('{?')
    else:
      wx= ox+2
      while wx<cx:
        if v[wx] in _ws:
          break
        wx= wx+1
      vx= wx
      while vx<cx:
        if v[vx] not in _ws:
          break
        vx= vx+1
      parts.append((v[ox+2:wx], v[vx:cx]))
    ix= cx+2
  return parts


def checkFeatures(future):
  """ Convert a future attribute node into a list of optional extra features.
      Raise an exception for any features that cannot be used in the current
      version of Python.
  """
  if future==None:
    return []
  features= map(string.strip, string.split(future.value, ','))
  for feature in features[:]:

    # If the feature is not implemented at all in the current version of
    # Python, complain. If the feature is on by default anyway, don't bother
    # including it in the import list.
    #
    if hasattr(future__mod, feature):
      f= getattr(future__mod, feature)
      if sys.version_info>=f.getOptionalRelease():
        if sys.version_info>=f.getMandatoryRelease():
          features.remove(feature)
        continue
    raise PXTLFutureFeatureError(future, feature)
  return features

def haveNestedScopes(features):
  """ Determine whether PXTL should provide nested scopes to subtemplate
      functions. Check the supplied features string and the current Python
      version's default features.
  """
  if 'nested_scopes' in features:
    return True
  if hasattr(future__mod, 'nested_scopes'):
    if sys.version_info>=future__mod.nested_scopes.getMandatoryRelease():
      return True
  return False

class ConditionalContext:
  """ Keeps track of previous clauses of an if-orif-anif-elif-else construct
      and calculates the success of each element. Used by conditional elements
      and conditional attributes.
  """
  def reset(self):
    """ Start a new conditional context, forgetting any previous clauses. This
        also serves as the constructor.
    """
    self.old_cond= None
    self.old_testing= None
    self.old_success= None
  __init__= reset

  def getTesting(self, cond):
    """ Given the new clause 'cond' and knowledge of previous clauses, decide
        whether the clause's test condition needs to be evaluated to determine
        its success. If not, laziness kicks in and we can avoid doing the
        test. getTesting for 'else' may return True even though 'else' doesn't
        actually have a clause to evaluate. If the clause ordering makes no
        sense (eg. if-else-orif), return None instead of a boolean.
    """
    if cond=='if':
      return True
    elif cond=='orif':
      if self.old_cond in ('if', 'orif'):
        return not self.old_success
    elif cond=='anif':
      if self.old_cond in ('if', 'orif', 'anif', 'for', 'while'):
        return self.old_success
    elif cond in ('elif', 'else'):
      if self.old_cond in ('if', 'orif', 'anif', 'for', 'while'):
        return not self.old_success
      if self.old_cond=='elif':
        return not self.old_success and self.old_testing
    return None

  def getSuccess(self, cond, test):
    """ Work out whether a clause is successful, given the truth of its
        evaluated test condition. If the condition was not evaluated (due to
        getTesting having returned False), 'test' should be None instead of a
        boolean. Store the clause info as it may affect a following clause.
    """
    if cond=='orif':
      success= test in (None, True)
    else:
      success= test
    self.old_cond= cond
    self.old_testing= test!=None
    self.old_success= success
    return success

  def copy(self):
    """ Make a copy of a ConditionalContext. Used for conditional attributes
        which can be hierarchically nested.
    """
    clone= ConditionalContext()
    clone.old_cond= self.old_cond
    clone.old_testing= self.old_testing
    clone.old_success= self.old_success
    return clone

class AbstractPXTLObject:
  """ Object passed into all template global scopes for output-writing and
      constant-holding purposes.
  """
  version= 1,-1
  pxtl= pxtl

  TEXT=    ('text/plain',             'text', None, None)
  CSS=     ('text/css',               'text', None, None)
  JS=      ('text/javascript',        'text', None, None)
  XML=     ('text/xml',               'xml',  None, None)
  TAGSOUP= ('text/html',              'html',
              '-//W3C//DTD HTML 4.01 Transitional//EN', None)
  HTML2=   ('text/html',              'html',
              '-//IETF//DTD HTML 2.0//EN', None)
  HTML3=   ('text/html',              'html',
              '-//W3C//DTD HTML 3.2 Final//EN', None)
  HTML4S=  ('text/html',              'html',
              '-//W3C//DTD HTML 4.01//EN',
              'http://www.w3.org/TR/html4/strict.dtd')
  HTML4T=  ('text/html',              'html',
              '-//W3C//DTD HTML 4.01 Transitional//EN',
              'http://www.w3.org/TR/html4/loose.dtd')
  HTML4F=  ('text/html',              'html',
              '-//W3C//DTD HTML 4.01 Frameset//EN',
              'http://www.w3.org/TR/html4/frameset.dtd')
  XHTML1S= ('text/html',              'xhtml',
              '-//W3C//DTD XHTML 1.0 Strict//EN',
              'http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd')
  XHTML1T= ('text/html',              'xhtml',
              '-//W3C//DTD XHTML 1.0 Transitional//EN',
              'http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd')
  XHTML1F= ('text/html',              'xhtml',
              '-//W3C//DTD XHTML 1.0 Frameset//EN',
              'http://www.w3.org/TR/xhtml1/DTD/xhtml1-frameset.dtd')
  XHTML11= ('application/html+xml',   'xhtml',
              '-//W3C//DTD XHTML 1.1//EN',
              'http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd')
  XHTMLB1= ('application/html+xml',   'xml',
              '-//W3C//DTD XHTML Basic 1.0//EN',
              'http://www.w3.org/TR/xhtml-basic/xhtml-basic10.dtd')
  MATH1=   ('application/mathml+xml', 'xml',
              None, 'http://www.w3.org/Math/DTD/mathml1/mathml.dtd')
  MATH2=   ('application/mathml+xml', 'xml',
              '-//W3C//DTD MathML 2.0//EN',
              'http://www.w3.org/TR/MathML2/dtd/mathml2.dtd')
  SVG1=    ('image/svg+xml',          'xml',
              '-//W3C//DTD SVG 1.0//EN',
              'http://www.w3.org/TR/2001/REC-SVG-20010904/DTD/svg10.dtd')
  SVG11=   ('image/svg+xml',          'xml',
              '-//W3C//DTD SVG 1.1//EN',
              'http://www.w3.org/Graphics/SVG/1.1/DTD/svg11.dtd')
  SVG11B=  ('image/svg+xml',          'xml',
              '-//W3C//DTD SVG 1.1 Basic//EN',
              'http://www.w3.org/Graphics/SVG/1.1/DTD/svg11-basic.dtd')
  SVG11T=  ('image/svg+xml',          'xml',
              '-//W3C//DTD SVG 1.1 Tiny//EN',
              'http://www.w3.org/Graphics/SVG/1.1/DTD/svg11-tiny.dtd')
  XHMS=    ('application/html+xml',   'xml',
              '-//W3C//DTD XHTML 1.1 plus MathML 2.0 plus SVG 1.1//EN', 
              'http://www.w3.org/2002/04/xhtml-math-svg/xhtml-math-svg.dtd')
